/*
    DU32 -- Shows disk Usage.  Similar to Unix DF command
    Copyright (C) 2005  Norm London, London Computing

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	norm_london@londoncomputing.com

    Compile with Visual Studio 6 or later or gcc 2.95.2 or later.
	gcc du32.c -o du32

*/

#include <stdio.h>
#include <time.h>
#include <process.h>
#include <string.h>
#include <stdlib.h>
// #include <graph.h>
#include <windows.h>
#include <conio.h>
#include <direct.h>

#define LABEL_LEN 32
char	label[LABEL_LEN];
#define ULONG unsigned long
#define USHORT unsigned short
DWORD SectorsPerCluster, 
	BytesPerSector, 
	NumberOfFreeClusters,
	TotalNumberOfClusters;

double avail[36] = {
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
double dsize[36] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int dtype[36] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
char dtypec[36][12];
char driveletters[]="@ABCDEFGHIJKLMNOPQRSTUVWXYZ";
char checkdrive[36] = {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
int dpct[36] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

double total_used = 0., total_avail=0., total_size = 0.;
char volume_label[32][32];
int first_time = 1;
long interval = 10000;
DWORD DriveBitMap=0;
char AllDrives[256];
char quiet=0;

char drivepath[4] = "C:\\" ;

void setDriveType() {
	int i;

	for (i=1;i<27;++i) {
		// If drive not specified on command line, don't bother checking it
		if (checkdrive[i] == 0) {
			avail[i] = -1;
			continue;
		}

		drivepath[0] = driveletters[i];
		dtype[i] = GetDriveType(drivepath);
		switch(dtype[i]) {
			case DRIVE_UNKNOWN: strcpy(dtypec[i],"Unknown"); break;
			case DRIVE_REMOTE: strcpy(dtypec[i],"Network"); break;
			case DRIVE_REMOVABLE: strcpy(dtypec[i],"Removable"); break;
			case DRIVE_NO_ROOT_DIR: 
				strcpy(dtypec[i],"Unmounted"); 
				avail[i] = -1;
				break;
			case DRIVE_FIXED: strcpy(dtypec[i],"Fixed"); break;
			case DRIVE_CDROM: strcpy(dtypec[i],"CDROM"); break;
			case DRIVE_RAMDISK: strcpy(dtypec[i],"RAMDISK"); break;
		}
	}
	DriveBitMap = GetLogicalDrives();
	GetLogicalDriveStrings(sizeof(AllDrives),AllDrives);
	return;
}

int IsNetwork(int drive) {
	drivepath[0] = '@'+drive;
	return (GetDriveType(drivepath) == DRIVE_REMOTE ? 1 : 0);
}

int IsRemovable(int drive) {
	drivepath[0] = '@' + drive;
	return (GetDriveType(drivepath) == DRIVE_REMOVABLE ? 1 : 0);
}

void beep(void) {
	putchar(7);
}

void Usage() 	
{
	FILE *hlp;
	char tmp[_MAX_PATH];
	char cmd[_MAX_PATH];
	char *wp;
	
	if ((wp=getenv("tmp")) == NULL)
		if ((wp = getenv("temp")) == NULL)
			strcpy(tmp,"c:");
		else
			strcpy(tmp,wp);
	else
		strcpy(tmp,wp);
	if (tmp[strlen(tmp)-1] != '\\')
		strcat(tmp,"\\");

	strcat(tmp,"du32.hlp");
	hlp = fopen(tmp,"w");

	
	fprintf(hlp,"DU32 %s, Copyright 2005, Norm London, London Computing.\n\n",__DATE__);
	fprintf(hlp,"This program is free software; you can redistribute it and/or modify\n");
	fprintf(hlp,"it under the terms of the GNU General Public License as published by\n");
	fprintf(hlp,"the Free Software Foundation; either version 2 of the License, or\n");
	fprintf(hlp,"(at your option) any later version.\n\n");
	fprintf(hlp,"This program is distributed in the hope that it will be useful,\n");
	fprintf(hlp,"but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
	fprintf(hlp,"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
	fprintf(hlp,"GNU General Public License for more details.\n\n");
	fprintf(hlp,"You should have received a copy of the GNU General Public License\n");
	fprintf(hlp,"along with this program; if not, write to the Free Software\n");
	fprintf(hlp,"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n");
	fprintf(hlp,"Written by London Computing, PO Box 696, Cherry Hill, NJ 08003.\n\n");
	fprintf(hlp,"Usage is: du32 /? or du [-v] [{-K|-M}] [-l[logfile]] [d: [pp%%]] ...\n");
	fprintf(hlp,"Displays spaced used, available for 1 or more drives\n");
	fprintf(hlp,"-v   -- Display unique volumes only.  Do not display multiple network\n");
	fprintf(hlp,"        drives that map to the same Network Volume.\n");
	fprintf(hlp,"-c   -- Continuous display marking drives as free space changes.\n");
	fprintf(hlp,"        In conjunction with pp%% option, can be used to continuously \n");
	fprintf(hlp,"        monitor one or more drivers.\n");
	fprintf(hlp,"-k   -- Display space in KB\n");
	fprintf(hlp,"-m   -- Display space in MB\n");
	fprintf(hlp,"-t   -- Total Disk Size, Used Space and Free Space\n");
	fprintf(hlp,"-l[logfile] -- log changes in diskspace in file logfile.\n");
	fprintf(hlp,"               If logfile not specified, write to dulog$$$.log\n\n");
	fprintf(hlp,"d:   -- specifies a drive letter to display. \n");
	fprintf(hlp,"pp%% -- is a desired percentage of space available for the drive.\n");
	fprintf(hlp,"        If a percentage is specified, but that amount of space\n");
	fprintf(hlp,"        is not available, an asterisk is printed next to the drive\n");
	fprintf(hlp,"        letter and a beep is sounded.  pp%% is optional.\n");
	fprintf(hlp,"You may place more than one set of drive letters and optional free\n");
	fprintf(hlp,"space percentages on the command line, e.g.\n");
	fprintf(hlp,"DU -LX:SPACELOG.LOG C: 20%% D: E: 15%\n\n");
	fprintf(hlp,"DU will log changes in space available to x:spacelog.log\n");
	fprintf(hlp,"Only drives C, D and E will be displayed.  Drive C will be flagged\n");
	fprintf(hlp,"unless there is at least 20%% free space.  Drive D will not be flagged.\n");
	fprintf(hlp,"Drive E will be flagged unless there is at least 15%% free space.\n\n");
	fprintf(hlp,"/? -- display this help.\n");
	fclose(hlp);
	strcpy(cmd,"NOTEPAD.EXE ");
	strcat(cmd,tmp);
	WinExec(cmd,SW_SHOW);
	exit(1);
}

char *comma_delim(double lno,char *buffer,char mb,char kb) {
/* returns the long integer in comma delimited format */
/* Assume buffer is long enough to hold the largest value */
	char b[36]; /* should be long enough to hold largest ltoa value */
	int i,n,l,k,c;
	double no;
	no = lno;
    if (mb) {
        no = (lno / (1024*1024));
        if (no == 0) {
            kb = 1;
            mb = 0;
        }
    }
    if (kb) {
        no = (lno / 1024);
        if (no == 0) {
            kb = 0;
            no = lno;
        }
    }
    sprintf(b,"%.0f",no);
	l = strlen(b); /* l is length of string and index to 0 at end of string */
	n = l/3;	   /* n is number of commas to be inserted */
	if ((l % 3) == 0)
		--n;
	k = l + n;	   /* k is length of new string */
	buffer[k--] = b[l--];
	for (c=0,i=l; i >= 0; --i,--k) {
		buffer[k] = b[i];
		if (++c == 3 && i > 0) {
			buffer[--k] = ',';
			c = 0;
		}
	}
    if (mb)
        strcat(buffer,"MB");
    else if (kb)
        strcat(buffer,"KB");
	return buffer;
}

#define FILESYSTEMNAME_LEN 32

char logfile[_MAX_DIR];



int main(int argc,char *argv[]) {
	DWORD VolumeSerialNumber, MaximumComponentLength, FileSystemFlags;
	char FileSystemName[FILESYSTEMNAME_LEN];
	double drive_size, drive_avail;
	int pct_avail;
	ULONG LogicalDrives;
	USHORT DriveNumber = 0, i;
	int skip,k,shortform = 0,c;
	int pct=0; /* if %available < this value, beep */
    char du_size[15], du_avail[15],du_used[15], updown;

	char temp[24],temp1[24],kb=0,mb=0,cont=0,more=1,totals=0;
	int rc;
	FILE *dulog;
	char log = 0;

	time_t start_time;

	strcpy(logfile,"DULOG$$$.LOG");
			/* Usage is du [-l[logfile]] [d: [pct]] ... | du /? */

	if (argc > 1 && strcmp(argv[1],"/?") == 0) 
		Usage();
	

	fprintf(stderr,"DU32 %s, Copyright 2005, Norm London, London Computing.\n",__DATE__);
	fprintf(stderr,"Enter DU32 /? to Display help.\n");


	// process switches and drive letters
	for (i=1;i < argc;++i) {
		if (argv[i][0]== '-') {
			switch(toupper(argv[i][1])) {
				case 'V': shortform = 1; break;
				case 'T': totals = 1; break;
				case 'C': cont= 1; break;
				case 'K': kb = 1; break;
				case 'M': mb = 1; break;
				case 'L': 
					log = 1;
					if (argv[i][2])
						strcpy(logfile,argv[i]+2);
					break;
			}
		}
		else {
			int j;
		
			if (isalpha(argv[i][0])) {
				if (checkdrive[0]) {
					// specified at least on drive to check and the checkdrive
					// array has not been reset, reset it to all 0s
					for (j=0;j<27;checkdrive[j++]=0);
				}
				// set drive to check
				checkdrive[toupper(argv[i][0])-'@'] = 1;
				// check next parm.  If numeric, set pct indicator for this drive
				if (i+1 < argc) {
					j = atoi(argv[i+1]);
					dpct[toupper(argv[i][0])-'@'] = j;
				}
			}
			else {
				fprintf(stderr,"%s -- Invalid Argument\n");
			}
		}
	}


	LogicalDrives = 0xFFFFFFFC;
	setDriveType();
	while (more) {	
		more = cont;
		total_size = total_avail = total_used = 0.0;
	//	_settextposition(0,0);
        puts("\nVol  Volume Label      Volume Size        Space Used         Available   %Avl");
        puts("---  ------------      -----------      ------------      ------------   ------");
		for (i = 3;i <= 26;++i) {
			if (checkdrive[i] == 0)
				continue;
			if (avail[i] == -1)
				continue;
			// check to see if drive is mounted (only necessary for CDROM and Removable)
			if (dtype[i] == DRIVE_REMOVABLE || dtype[i] == DRIVE_CDROM || dtype[i] == DRIVE_UNKNOWN) {
				if (_chdrive(i)) {
					// Unable to change the drive, mark it unavailabe
					avail[i] = -1;
					continue;
				}
			}

			drivepath[0] = '@' + i;
			rc = GetDiskFreeSpace(drivepath,&SectorsPerCluster,
				&BytesPerSector,&NumberOfFreeClusters,&TotalNumberOfClusters);
			if (rc == TRUE) {
				strcpy(label,"                               ");
				rc =GetVolumeInformation(drivepath,label,LABEL_LEN,
					&VolumeSerialNumber,&MaximumComponentLength,
					&FileSystemFlags,FileSystemName,FILESYSTEMNAME_LEN);
				// set indicator showing whether space greater or less than
				// last time checked (for continuous operation only)
				if (first_time)
					updown = ' ';
				else
					if (avail[i-1] == NumberOfFreeClusters)
						updown = ' ';
					else
						if (avail[i-1] < NumberOfFreeClusters)
							updown = 30;
						else
							updown = 31;
				// Gather space stats
				avail[i-1] = NumberOfFreeClusters;
				strcpy(volume_label[i-1],label);
				drive_size = (double)SectorsPerCluster * 
							(double)TotalNumberOfClusters * 
							(double)BytesPerSector;
				drive_avail= (double)SectorsPerCluster * 
							(double)NumberOfFreeClusters * 
							(double)BytesPerSector;
				pct_avail = (int)(drive_avail*100. / drive_size);
				dsize[i-1] = drive_size;			
				skip = 0;

				// If V flag set, look to see if this volume was already reported
				if (shortform) {
					for (skip=0,k=1;k<i && !skip;++k) {
						// set skip if drive already there
						if (avail[k] != -1 && dtype[k]== DRIVE_REMOTE) {
							if (avail[i] == avail[k] || dsize[i] == dsize[k]) {
								if (strcmp(volume_label[k],label) == 0) {
									skip = 1;
									avail[i] = -1;
								}
							}
						}
					}
				}

				if (!skip) {
					// If flag is not set, print drive space stats
                    printf("%c%c:%c %-12s  %15s   %15s   %15s%c   %d%%\n",
					((int)(drive_avail * 100 / drive_size) < dpct[i] ? '*' : ' '),
					'@'+i,
					updown,
					label,
					comma_delim(drive_size,du_size,mb,kb),
					comma_delim(drive_size - drive_avail,du_used,mb,kb),
					comma_delim(drive_avail,du_avail,mb,kb),
					dtype[i]==DRIVE_REMOTE ? '*' : ' ',pct_avail);
					
					total_size += drive_size;
					total_avail += drive_avail;
					total_used += (drive_size - drive_avail);

					if ((int)(drive_avail * 100 / drive_size) < dpct[i])
						beep();
				}
				if (log && (first_time || updown != ' ')) {
					dulog = fopen(logfile,"a");
					if (dulog != NULL) {
						fprintf(dulog,"%s %s %c%c: %-12s %12ld %12ld\n",
						_strdate(temp),
						_strtime(temp1),
						((int)(drive_avail * 100 / drive_size) < pct ? '*' : ' '),
						'@'+i,
						label,
						drive_size,
						drive_avail);
						fclose(dulog);
					}
				}
			}
			else
				avail[i] = -1l;
		}
		

		if (totals) 
			printf("---\nTotal:             %15s   %15s   %15s%c\n",
			comma_delim(total_size,du_size,mb,kb),
			comma_delim(total_used,du_used,mb,kb),
			comma_delim(total_avail,du_avail,mb,kb),
			' ');
		
		first_time = 0;
		
		if (more) {
			printf("\nScan interval: %2ld seconds, + (faster), - (slower) ESC (Quit)",
				interval/1000);
		
			for (i=0;i < (int)(interval/1000);++i) {
				start_time = time(&start_time);
				Sleep(1000);
				if (kbhit()) {
					c = getch();
					switch(c) {
					case 27:
						exit(0);
					case '+':
						if (interval >= 4000)
							interval -= 2000;
						else
							interval = 2000;
						break;
					case '-':
						if (interval <= 58000)
							interval += 2000;
						else
							interval = 60000;
						break;
					}
				}
			}
		}
		
	}
	return 0;
}
